import os 
import re
import platform


# toutes les fonctions utilisees par les etudiants commencent par verifier le type des parametres reels 

################ PRIMITIVES GRAPHE   #################################

# retourne la liste des sommets du graphe G
def liste_sommets(G):
    if not verif_type_graphe(G):
        raise ErreurParametre("liste_sommets","un graphe")
    else:
        return G.nodes

# retourne l'etiquette du graphe G
def nom_graphe(G):
    if not verif_type_graphe(G):
        raise ErreurParametre("nom_graphe","un graphe")
    else:
        return _label(G)

# retourne le nombre de sommets du graphe G
def nb_sommets (G):
    if not verif_type_graphe(G):
        raise ErreurParametre("nb_sommets","un graphe")
    else:
        return len(liste_sommets(G))

def sommet(G,i):
    if not verif_type_graphe(G) or not verif_type_entier(i) or i < 0 or i > nb_sommets(G)-1:
        raise ErreurParametre("sommet","un graphe G", "un entier compris entre 0 et nb_sommets(G) -1")
    else:
        return liste_sommets(G)[i]

################# PRIMITIVES SOMMET   ###################################

# retourne l'etiquette associee au sommet s
def nom_sommet(s):
    if not verif_type_sommet(s):
        raise ErreurParametre("nom_sommet","un  sommet")
    else:
        return str(s.label)

# retourne le degre du sommet s
def degre (s) :
    if not verif_type_sommet(s):
        raise ErreurParametre("degre","un sommet")
    else:
        return len(liste_aretes_incidentes(s)) 

# retourne la liste des aretes incidentes au sommet s
def liste_aretes_incidentes(s):
    if not verif_type_sommet(s):
        raise ErreurParametre("liste_aretes_incidentes","un sommet")
    else:
        return s.edges


# retourne l'autre extremite de l'arete a incidente a s 
# ne retourne rien si a n'est pas incidente a s
# FAUT-IL GENERER UN MESSAGE d'ERREUR ?
def sommet_voisin_par(s, a):
    if not verif_type_sommet(s) or not verif_type_arete(a):
        raise ErreurParametres("sommet_voisin_par","un sommet", "une arete")
    else:
        if a.start == s:
            return a.end
        elif a.end == s:
            return a.start
        else:
            return None

def voisin(s, i):
    if not verif_type_sommet(s) or not verif_type_entier(i) or i < 0 or i > degre(s):
        raise ErreurParametres("voisin", "un sommet s", "un entier compris entre 0 et degre(s)-1")
    else:
        return sommet_voisin_par(s, liste_aretes_incidentes(s)[i])

# pose une marque sur le sommet s
def marquer_sommet (s):
    if not verif_type_sommet(s):
        raise ErreurParametre("marquer_sommet","un sommet")
    else:
        s.mark = True

# supprime la marque eventuelle du sommet s
def demarquer_sommet (s):
    if not verif_type_sommet(s):
        raise ErreurParametre("demarquer_sommet","un sommet")
    else:
        s.mark = False

# retourne True si le sommet s est marque False sinon
def est_marque (s):
    if not verif_type_sommet(s):
        raise ErreurParametre("est_marque","un sommet")
    else:
        return s.mark

# attribue la couleur color au sommet s
def colorier(s,color):
    if not verif_type_sommet(s) or not verif_type_chaine(color):
        raise ErreurParametres("colorier_sommet","un sommet", "une couleur")
    else:   
        s.color = str(color)

# retourne la couleur du sommet s
def couleur(s):
    if not verif_type_sommet(s):
        raise ErreurParametre("couleur_sommet","un sommet")
    else:
        return str(s.color)
    

################# FONCTIONS SOMMET   #################################

# retourne la liste des sommets voisins de s
def liste_voisins(s):
    if not verif_type_sommet(s):
        raise ErreurParametre("liste_voisins","un sommet")
    else:
        voisins=[]
        for e in liste_aretes_incidentes(s):
            voisins.append( sommet_voisin_par(s,e))
        return voisins

# retourne le sommet d'etiquette etiquette dans le graphe G
# ne retourne rien si un tel sommet n'existe pas 
def chercher_sommet(G,etiquette):
    if not verif_type_graphe(G) or not verif_type_chaine(etiquette):
        raise ErreurParametres("chercher_sommet","un graphe","une etiquette")
    else:
        for s in liste_sommets(G):
            if nom_sommet(s) == etiquette:
                return s
        return None


################ PRIMITIVES ARETE ########################

# retourne l'etiquette associee a l'arete a
def nom_arete(a):
    if not verif_type_arete (a):
        raise ErreurParametre("nom_arete","une arete")
    else:
        return str(a.label)

# pose une marque sur l'arete a
def marquer_arete (a):
    if not verif_type_arete (a):
        raise ErreurParametre("marquer_arete","une arete")
    else:
        a.mark = True

# supprime la marque eventuelle de l'arete a
def demarquer_arete (a):
    if not verif_type_arete (a):
        raise ErreurParametre("demarquer_arete","une arete")
    else:
        a.mark = False

# retourne True si l'arete a est marquee, False sinon
def tester_arete_marquee (a):
    if not verif_type_arete (a):
        raise ErreurParametre("tester_arete_marquee","une arete")
    else:
        return a.mark


############### fonctions d'affichage du graphe  #####################

# affiche une description du graphe G 
def afficher_graphe(G):
    if not verif_type_graphe(G):
        raise ErreurParametre("afficher_graphe","un graphe")
    else:
        print "description du graphe ",_label(G)
        print len(G.nodes) , "Sommet(s) :"
        for s in G.nodes:
            print "Sommets accessible depuis", _label(s),
            print "(",s.color,"):"
            for a in s.edges:
                print '\t'," --", _true_label(a), "--",
                print _label(_alter_node_edge(s, a))
        return None

def afficher(G):
    afficher_graphe(G)

# redessine le graphe G
def redessiner_graphe(G):
    if not verif_type_graphe(G):
        raise ErreurParametre("redessiner_graphe","un graphe")
    else:
        dessiner_graphe(G,'dot',True,'Black',False)
        return None

def redessiner(G):
    redessiner_graphe(G)

# Dessine une representation graphique du graphe G
# La valeur 'png' pour le parametre commande cree un fichier png
# Sous OS X, la visualisation necessite graphviz.
# Sous Linux, elle necessite graphviz et gv.
def dessiner_graphe(G, commande = 'dot', disp_edge_label = True, colormark = 'Black', launch_gv = True):
    if not verif_type_graphe(G):
        raise ErreurParametre("dessiner_graphe","un graphe")
    else:
        _unmark_all_edges(G)

        if _label(G) == '':
            nom_graphe = 'G'
        else:
            nom_graphe = _label(G)

        graph_dot = nom_graphe + ".dot"
        
        f = os.open(graph_dot,os.O_CREAT|os.O_TRUNC|os.O_RDWR)
        os.write(f, "graph " + nom_graphe +"{\n" + G.drawopts + "\n")

        for s in G.nodes:
            d = len(s.edges)            
            if d== 0:
                os.write(f," \"" + _label(s) + "\";\n")
            for a in s.edges:
                if not _marked(a):
                    _mark(a)
                    os.write(f, "  \"" + _label(s) + "\" -- \"" +
                         _label(_alter_node_edge(s,a))+"\"")
                    if disp_edge_label:
                        os.write(f, " [ label = \"" + _true_label(a) + "\" ];\n")
                    else:
                        os.write(f, " ;\n")
            for s in G.nodes:
                color = "black"
                if s.mark:
                    entoure=2
                    color = colormark
                else:
                    entoure=1
                fillcolor = s.color
                if fillcolor <> 'white':
                    os.write(f,"\"" + _label(s) + "\"" + " [style = filled,peripheries=" +
                             str(entoure) + ", fillcolor = " + s.color +", color = " + color +"];\n");
                elif s.mark:
                    os.write(f,"\"" + _label(s) +"\""  + " [peripheries = 2" + ", color = " + color + "];\n");
            
        os.write(f, "}\n")
        os.close(f)

        _unmark_all_edges(G)
    
        if commande == "dot":
            graph_ps = nom_graphe + ".ps"
            os.spawnlp(os.P_WAIT, 'dot', 'dot', '-Tps', graph_dot, '-o', graph_ps)
            if launch_gv:
                if re.match("^Darwin", platform.platform()):
#                    os.spawnlp(os.P_NOWAIT, 'open', 'open', graph_ps)
                    os.spawnlp(os.P_NOWAIT, 'open', 'open', graph_dot)
                else:
                    os.spawnlp(os.P_NOWAIT, 'kghostview', 'kghostview','-watch', graph_ps)
            elif commande == "dotty":
                os.spawnlp(os.P_NOWAIT, 'dotty', 'dotty', graph_dot)
        elif commande == "png":
            graph_png = nom_graphe + ".png"
            os.spawnlp(os.P_WAIT, 'dot', 'dot', '-Tpng',graph_dot, '-o', graph_png)
        

def dessiner(G, commande = 'dot', disp_edge_label = True, colormark = 'Black', launch_gv = True):
    dessiner_graphe(G, commande, disp_edge_label, colormark, launch_gv)

        
#### Fonctions de creation du graphe

# Colore les sommets du graphe G a partir de la liste colors
# chaque element de la liste contient une liste 
# etiquette du sommet, couleur a affecter au sommet
def color_graph(G,colors):
    for i in range(len(colors)):
        s=_search_node(G.nodes,colors[i][0])
        s.color=str(colors[i][1])
    return None

# Cree un graphe
# paths est la description du graphe
# label est l'etiquette du graphe
# colors contient les couleurs des sommets
# drawopts est une chaine passee au programme d'affichage dot (devrait etre ailleurs qu'ici).
def make_graph (paths, label='',colors = [], drawopts =""):
    G = c_graph(label)
    _fill_graph(G, paths)
    if len(colors)<> 0:
        color_graph(G,colors)
    G.drawopts = drawopts
    return G


##### Gestion des erreurs

def verif_type_graphe(G):
    return type(G)<> type(1) and type(G)<>type(' ') and G.nature == 'Graphe'
    
def verif_type_sommet(s):
    return type(s)<> type(1) and type(s)<>type(' ') and s.nature == 'Sommet'

def verif_type_chaine(s):
    return type(s) == type(' ')

def verif_type_entier(s):
    return type(s) == type(1)

def verif_type_arete(a):
    return type(a)<> type(1) and type(a)<>type(' ') and a.nature == 'Arete'

def erreur(message):
    message_erreur = "\n****************     Attention Erreur    **********************\n"
    message_erreur += "*"
    message_erreur += message
    message_erreur +="\n***************************************************************"
    return message_erreur

def erreur_parametre(nom_fonction,nom_parametre,parametre_reel ):
    return erreur("L'argument attendu par "+str(nom_fonction) +" est " + nom_parametre + " et non " + parametre_reel)

   
class ErreurParametre(Exception):
    def __init__(self,nom_fonction,nom_parametre):
        self.nom_fonction = nom_fonction
        self.nom_parametre = nom_parametre
    def __str__(self):
            return  "\n\n****************     Attention Erreur    **********************\n*" + \
                   " L'argument attendu par " + self.nom_fonction + " est " + self.nom_parametre + \
                   "\n***************************************************************"

class ErreurParametres(Exception):
    def __init__(self,nom_fonction,nom_parametre1,nom_parametre2):
        self.nom_fonction = nom_fonction
        self.nom_parametre1 = nom_parametre1
        self.nom_parametre2 = nom_parametre2

    def __str__(self):
            return  "\n\n****************     Attention Erreur    **********************\n*" + \
                   " Les arguments attendus par " + self.nom_fonction + " sont " + self.nom_parametre1 + \
                   " et " + self.nom_parametre2 + \
                   "\n***************************************************************"
    
# Definitions des classes 
# les _mark sont des marques internes non connues des utilisateurs 

class c_edge:
    "une arete du graphe"
    def __init__(self, label = '', start = None, end = None, mark = False):
        self.label = label
        self.start = start
        self.end = end
        self.mark = mark
        self._mark = False
        self.nature='Arete'
    def __repr__(self):
        return "<Arete " + str(self.label) \
               +" Marque : " + str(self.mark) + ">"

class c_node:
    "un sommet du graphe"
    def __init__(self, label = '', color = "white", mark = False):
        self.label = label
        self.color = color
        self.mark = mark
        self._mark = False
        self.edges = []
        self.nature = "Sommet"
    def __repr__(self):
        if self.color != "white":
            colorprint = str(self.color)
        else:
            colorprint = "-"

        if self.mark:
            markprint = "marque"
        else:
            markprint = "-"
        # python 2.5, commente pour compatibilite Agen
        # return "<" + str(self.label)+","+ (str(self.color)if self.color != "white"else "-")\ +","+("*" if self.mark else "-") + ">"
        return "<" + str(self.label) +"," + colorprint +","+ markprint +">"

class c_graph:
    "un graphe oriente sommets"
    def __init__(self, label = ''):
        self.nodes = []
        self.label = label
        self.nature = "Graphe"
        # devrait etre separe de la classe graphe, concerne l'affichage...
        self.drawopts = ""
    def __repr__(self):
        return "<Graphe "+ str(self.label) + " avec  " + str(len(self.nodes))\
               + " sommets>"

#Fonction internes utiles

def _label(object):
    return str(object.label)

def _change_label(object, newlabel):
    object.label = newlabel
    return str(object.label)

def _true_label(object):
    if str(object.label).startswith("internal_"):
        return ''
    return str(object.label)

def _mark(object):
    object._mark = True
    return None

def _unmark(object):
    object._mark = False
    return None

def _marked(object):
    return object._mark
 
def _unmark_all_edges(G):
    for s in G.nodes:
        for e in s.edges:            
            _unmark(e)
    return None

def _alter_node_edge(s,e):
    if e.start == s:
        return e.end
    elif e.end == s:
        return e.start
    return None

def _find_add_node (ln, nodes):
    i = _find_node(ln, nodes)
    if i == len(nodes):
        nodes.append(c_node(ln))
    return i

def _find_node (ln, nodes):
    j = len(nodes)
    for i in range( j):
        if (_label(nodes[i]) == ln):
            return i
    return j

def _search_node(nodes,label):
   for s in nodes:
        if _label(s) == label:
            return s
   return None

def _add_edge (G, label, i, j):
    a = c_edge(label,G.nodes[i],G.nodes[j])
    G.nodes[i].edges.append(a)
    G.nodes[j].edges.append(a)
    return a

def _fill_graph (G, paths):
    nba = 0
    for p in paths:
        labelsource = p[0]
        i = _find_add_node(labelsource, G.nodes);
        for k in range(1,len(p)):
            edge_with_label = type(p[k]) == type([])
            if edge_with_label:
                labeldestination = p[k][0]
                labeledge = p[k][1]
            else:
                labeldestination = p[k]
                labeledge = "internal_"+str(nba)
            j = _find_add_node(labeldestination, G.nodes)
            _add_edge(G, labeledge, i, j)
            nba = nba + 1
    return None


# gestion automatique des couleurs

from colors import *
